#!/usr/bin/tixwish4.1.8.0
#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: FormServer.tcl,v 1.14 2000/10/15 17:19:20 artur Exp $
#
# Information:
#
# Dies ist eine Startdatei fr ein FormularServer
# Das Programm liest die xmlform Datei (Beschreibung der DB Formulare als XML)
# verbindet sich mit Datenbank
# und setz die Operationen auf Formularen in SQL Anweisungen

# Finde Program-Verzeichnis
set sname [info script]
if {$sname==""} {
    # Run interactive for develop purposes
    set progdir /home/artur/diplom/program
} else {
    file lstat $sname stats
    # follow sym links
    if {$stats(type)=="link"} {
	set sname [file readlink $sname]
	if {[file pathtype $sname]=="relative"} {
	    set sname [file join [file dirname [info script]] $sname]
	}
    }
    set progdir [file dirname $sname]
}
wm withdraw .
# Lade Weitere Tcl Komponenten Xotc,tDom,TclX
source [file join $progdir Init.tcl]

#Lade andere Module
source [file join $progdir sql.tcl]
source [file join $progdir formengine.tcl]
source [file join $progdir formfactory.tcl]

#Setzte Array lan mit bersetzung-Strings
checklanguage fs
#Dom Schnittstelle auch auf deutsche Umlauten setzten
dom setResultEncoding iso8859-1

# Haupt-Objekt des Programm das die GUI aufbaut und kontroliert
# Nur ein per Ausfhrung (singleton)

Object FormServer
FormServer proc setwin win {
    [self] instvar window
    set window $win
    global lan
    menu $win.mb -tearoff 0
    menu $win.mb.file -tearoff 0
    menu $win.mb.view -tearoff 0
    menu $win.mb.help -tearoff 0
    
    $win.mb add cascade -label $lan(menu.file.file) -menu $win.mb.file \
	-underline 0
    $win.mb add cascade -label $lan(menu.help.help) -menu $win.mb.help \
	-underline 0
    
    $win.mb.file add command -label $lan(menu.file.readxmlforms) \
	-command {FormServer readxmlforms}
    $win.mb.file add command -label $lan(menu.file.dbconnect) \
	-command {FormServer dbconnect}
    $win.mb.file add command -label $lan(menu.file.dbdisconnect) \
	-command {FormServer dbdisconnect}
    $win.mb.file add separator
    $win.mb.file add command -label $lan(menu.file.exit) \
	-command {FormServer handle_exit}

    $win.mb.help add command -label $lan(menu.help.help) \
	-command handle_help
    $win.mb.help add command -label $lan(menu.help.about) \
	-command {handle_about FormServer}
    $win.mb.help add separator
    $win.mb.help add command -label $lan(menu.help.debug) \
	-command {commandloop}
    $win.mb.help add command -label inspector \
	-command {source [file join $progdir inspector.tcl]}
    $win.mb.help add command -label win_inspector \
	-command {source [file join $progdir win_inspector.tcl]}
    $win.mb.help add command -label "view XML Forms" \
	-command {
	    if {![winfo exist .xmlview]} {
		toplevel .xmlview
		tixScrolledText .xmlview.st
		pack  .xmlview.st -expand yes -fill both
	    }
	    .xmlview.st subwidget text delete 1.0 end
	    .xmlview.st subwidget text insert 1.0 [$FormServer::data_node asXML]
	}
    $win.mb.help add command -label "sql DEBUG" \
	-command {source [file join $progdir sql_debug.tcl]}
     
    tixScrolledTList $win.list  -options {
	tlist.orient horizontal
	tlist.selectMode single
	tlist.command {FormServer executeForm}
    }
    pack $win.list -expand yes -fill both 
    
    if {$win==""} {
	. configure -menu $win.mb
	wm protocol . WM_DELETE_WINDOW "[self] handle_exit"
    } else {
	$win configure -menu $win.mb
	wm protocol $win WM_DELETE_WINDOW "[self] handle_exit"
    }
}
FormServer proc readxmlforms {{file {}}} {
    [self] instvar window data_node dom_doc db_name sqlconnect form_nodelist
    global lan
    # !!! Suche andere Methode das Tree window zu finden
    set types {
	{"XML Forms"		{.xmlforms}	}
	{"All files"		*}
    }
    if {$file==""} {
	set file [tk_getOpenFile -filetypes $types -parent .]
	if {$file == ""} { return }
    }
    if {[info exist sdom_doc]} {
	set alt_dom_doc $dom_doc
    } else {
	set alt_dom_doc {}
    }
    set flist [$window.list subwidget tlist]
    set fileh [open $file]
    set dom_doc [dom parse [read $fileh]]
    close $fileh
    set data_node [$dom_doc documentElement]
    if {[$data_node nodeName]!="dbforms"} {
	$dom_doc delete
	set dom_doc ""
	tk_messageBox -message $lan(message.wrongxml) -icon error -type ok
	return
    }
    set db_name [$data_node getAttribute name]
    set dbconnect_par {}
    set form_nodelist {}
    $flist delete 0 end
    foreach node [$data_node childNodes] {
	if {[$node nodeType]=="ELEMENT_NODE" && [$node nodeName]=="form"} then {
	    $flist insert end -itemtype imagetext \
		-text [$node getAttribute name] \
		-image [tix getimage openfold]
	    lappend form_nodelist $node
	} elseif {[$node nodeType]=="PROCESSING_INSTRUCTION_NODE" && [$node target]=="dbconnect"} {
	    set dbconnect_par [$node data]
	}
    }
    # Alte Strukturen und SQL-Verbindung falls vorhanden Lschen
    [self] dbdisconnect
    wm title . [file tail $file]:$db_name
    set sqlconnect [ConnectCreator getconnect [Object autoname sqlconnect] $dbconnect_par]
    if {$sqlconnect==""} {
	tk_messageBox -message $lan(connect.noconnect) -type ok -icon warning
    }
    if {$alt_dom_doc!=""} {
	$alt_dom_doc delete
    }
}

FormServer proc executeForm num {
    [self] instvar window data_node form_nodelist sqlconnect  
    #set flist [$window.list subwidget tlist]
    #set name [$flist entrycget $num -text]
    #Work Arround find bringt do Maschine zum Absturz
    #set form_node [$data_node find name $name]
    set fname [Object autoname myFormEngine]
    FormEngine $fname -form_node [lindex $form_nodelist $num] \
	-sqlconnect $sqlconnect 
    $fname map
    $fname reload
}
FormServer proc dbconnect {} {
    [self] instvar sqlconnect
    global lan
    if {[info exist sqlconnect] && $sqlconnect!=""} {
	tk_messageBox -message $lan(dialog.isconnect) -type ok -icon info
	return
    }
    set sqlconnect [ConnectCreator getconnect [Object autoname sqlconnect] ""]
    if {$sqlconnect==""} {
	tk_messageBox -message $lan(connect.noconnect) -type ok -icon warning
    }

}
FormServer proc dbdisconnect {} {
    [self] instvar sqlconnect
    if {[info exist sqlconnect] && $sqlconnect!=""} {
	$sqlconnect destroy
	set sqlconnect ""
    }
}
FormServer proc handle_exit {} {
    global lan
    if {[tk_messageBox -message $lan(dialog.endmsg) -icon question -type okcancel]=="ok"} { exit }
}
FormServer setwin {}
# Die FormEngine mssen zuerst zerstr werden weil sie 
# beim delete noch die SQL-Connect brauchen
Object setExitHandler {
    foreach obj [FormEngine info instances] {
	if {[FormEngine info instances $obj]!=""} {
	    $obj destroy
	}
    }
}
#FormServer readxmlforms mydb2.xmlforms
#Auswertung von Programparameter
if {!$tcl_interactive && [llength $argv]==1} {
    FormServer readxmlforms $argv
}
wm deiconify .
