#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: formengine.tcl,v 1.9 2000/10/15 17:19:20 artur Exp $
#
# Information:
# Modul (benutzt von Form Server)
# Formengine ist zustndigt fr ein DB Formular
# Es veranlast das Formular an Die WidgetFactory zu bauen
# Baut das Fenster mit Navigations Schaltflchen
# Initialisiert das DBProxys. unten)
# bernimmt das Auslesen und Setzten von Formular-Dialog (Maske)
# Eigene Implementierung von Callback Handler

Class NotifyHandler
NotifyHandler instproc addNotify {notify proc} {
    [self] instvar notifyproc
    lappend notifyproc($notify) $proc
}
NotifyHandler instproc clearNotify {notify proc} {
    [self] instvar notifyproc
    if {[info exist notifyproc($notify)]} {
	if {[set id [lsearch $notifyproc($notify) $proc]]>0} {
	    if {[llength $notifyproc($notify)]==1} {
		unset $notifyproc($notify)
	    } else {
		set notifyproc($notify) [lreplace $notifyproc($notify) $id $id]
	    }
	}
    }
}
NotifyHandler instproc invokeNotify {notify args} {
    [self] instvar notifyproc
    if {[info exist notifyproc($notify)]} {
	foreach nproc $notifyproc($notify) {
	    eval $nproc $args
	}
    }
}

Class FormEngine -superclass {InputWidget NotifyHandler} -parameter {
    form_node
    sqlconnect 
    {nestedform 0}
    {oneobject 0}
    data
    label
    linkedForm
    {refattribute {}}
    {border 0}
}
FormEngine metadata add attributes
FormEngine metadata attributes {label form refattribute}
FormEngine instproc init args {
    [self] instvar win dialog_container form_node \
	act_tupel old_tupel tupel_count sqlconnect attrlist refattribute \
	notnull_data def_values linkedForm oneobject nestedform
    # Nested Form sucht selbst den Vater
    # Genau so bei InheritForm
    if {$nestedform || $oneobject} {
	for {set pf_engine [[self] info parent]} \
	    {$pf_engine!="" && $pf_engine!="::" && [$pf_engine info class]!="::FormEngine"} \
	    {set pf_engine [$pf_engine info parent]} {}
	set linkedForm $pf_engine
    }
    if {[info exist linkedForm] && [FormEngine info instances $linkedForm]!=""} {
	# berprfung von Rekursiven Form Aggregation
	if {$nestedform && $form_node==[$linkedForm set form_node]} {
	    global lan
	    tk_messageBox -message $lan(message.rekforms) -type ok -icon error
	    error "XML Forms corrupt"
	}
	$linkedForm addNotify create "[self] createNewTupel"
	# Hier kann man auch ein Kaskadierendes Delete Aufbauen"
	if {$oneobject} {
	    $linkedForm addNotify delete "[self] delete"
	}
    }
    #puts "FormEngine Init $form_node $sqlconnect"
    set mf [FormFactory [Object autoname myformfactory] -sqlconnect $sqlconnect -hidedata $refattribute]
    set dialog_container [$mf makedialog_fromdom $form_node [self]]
    # Transferiere Datenherkunft Informationen
    set attrlist [$mf set data]
    set notnull_data [$mf set notnull_data]
    set def_values [$mf set def_values]
    $mf destroy
    #puts "DATA $attrlist"
    regexp {.*::(.+)} [self] _ win
    set win .$win
    #[self] build_window
    set table [$form_node getAttribute table]
    set primarykey [$form_node getAttribute primarykey]
    if {$sqlconnect!=""} {
	DBProxy [self]::myDBProxy \
	    -table $table -primarykey $primarykey -attrlist $attrlist \
	    -sqlconnect $sqlconnect -refattribute $refattribute
    } 
    set act_tupel 0
    set old_tupel -1
   #[self] reload
}
FormEngine instproc map {{window {}}} {
    [self] instvar form_node win dialog_container nestedform label data labelwin refattribute oneobject border
    if {$window!=""} {set win $window}
    #puts "WINDOW $win [self] $oneobject $nestedform"
    global lan
    if {!$nestedform} {
	#puts "not nested [self]"
	toplevel $win
	wm withdraw $win
	if {$refattribute==""} {
	    wm title $win [$form_node getAttribute name]
	} else {
	    wm title $win "[$form_node getAttribute name] Linked ($data->$refattribute)"
	}
	$dialog_container map $win.dialog
	
	menu $win.mb -tearoff 0
	menu $win.mb.file -tearoff 0
	$win.mb add cascade -label $lan(menu.file.form) -menu $win.mb.file \
	    -underline 0
	
	# list of {Name Image Command}
	pack [[self] buildNavigation $win.navigation] \
	    -side bottom -anchor s -expand yes -fill x
	pack $win.dialog -side top -expand yes -fill both
	$win configure -menu $win.mb 
	wm protocol $win WM_DELETE_WINDOW "[self] destroy"
	wm deiconify $win
    } else {
	#puts "nested [self]"
	if {[info exist label] && $label!=""} {
	    tixLabelFrame $win -label $label
	    set bwin [$win subwidget frame]
	} else {
	    if {$border==1} {
		frame $win -relief sunken -borderwidth 1
	    } else {
		frame $win
	    }
	    set bwin $win
	}
	$dialog_container map $bwin.dialog
	
	pack [[self] buildNavigation $bwin.navigation] \
	    -side bottom -anchor s -expand yes -fill x
	pack $bwin.dialog -side top -expand yes -fill both
    }
}
# Bildet die Naviagationleitst und vielleicht ein Men wenn
# Kein eingebeteter Folmular
# gibt ein Frame der Naviagationsleiste zurck
# Seiten Effect bei Bilden von Mens ($win.mb.file) 
#        als bereits initialisierte Men
FormEngine instproc buildNavigation {nwin} {
    [self] instvar win nestedform labelwin oneobject navwin sqlconnect
    global lan
    set navwin $nwin
    if {$sqlconnect==""} {
	::label $nwin -text $lan(connect.noconnectwar)
	return $nwin
    }
    frame $nwin  -bd 1 -relief ridge
    if {$oneobject} {
	# Name, Bitmap, Command
	set commands {
	    {save filefloppy updateManual}
	    {undo undo undo}
	    {delete delete delete}
	    {reload reload reload}
	}
    } else {
	set commands {
	    {first start first} 
	    {prev back prev}
	    {next forward next_tupel}
	    {last finish last}
	    {save filefloppy updateManual}
	    {undo undo undo}
	    {delete delete delete}
	    {reload reload reload}
	    {find find filterData}
	}
	tixLabelEntry $nwin.label -labelside right -label "von XX" \
	    -option {
		entry.width 4
	    }
	set labelwin $nwin.label
	pack $nwin.label -side left -anchor w
    }
    if {!$nestedform} {
	lappend commands {exit exit destroy}
    }
    foreach command $commands {
	set name [lindex $command 0]
	set image [lindex $command 1]
	set execute [lindex $command 2]
	if {!$nestedform} {
	    $win.mb.file add command -label $lan(navigation.${name}) \
		-command "[self] $execute"
	}
	button $nwin.$name -image [tix getimage $image] \
	    -command "[self] $execute" \
	    -takefocus 0 -bg #c0c0c0 -highlightbackground #c0c0c0
	pack $nwin.$name -side left -anchor w
    }
    return $nwin
}
#
FormEngine instproc stateNeedReload {} {
    [self] instvar navwin
    # Setzt den Hintegrund von Reload Button auf rot
    $navwin.reload configure -bg red
}
FormEngine instproc reload {} {
    [self] instvar labelwin tupel_count oneobject act_tupel navwin 
    if {[DBProxy info instances [self]::myDBProxy]==""} {
	return
    }
    set tupel_count [[self]::myDBProxy reload]
    #puts "RELOAD [self] $act_tupel $[self]::old_tupel"
    if {!$oneobject} {
	$labelwin subwidget label configure -text "von $tupel_count"
    } else {
	if {$tupel_count==0} {
	    $navwin.delete configure -state disabled
	} else {
	    $navwin.delete configure -state normal
	}
    }
    if {$tupel_count==0} {
	set act_tupel 0
    } elseif {$act_tupel>=$tupel_count} {
	set act_tupel [expr $tupel_count-1]
    }
    [self] fill_form
    $navwin.reload configure -bg #c0c0c0
    if {$oneobject} {
	$navwin.delete configure -state normal
	$navwin.reload configure -state normal
	$navwin.save configure -state normal
    } else {
	$navwin.first configure -state normal
	$navwin.prev configure -state normal
	$navwin.next configure -state normal
	$navwin.last configure -state normal
	$navwin.delete configure -state normal
	$navwin.reload configure -state normal
	$navwin.find configure -state normal
	$navwin.save configure -state normal
    }
}
FormEngine instproc fill_form {} {
    [self] instvar act_tupel old_tupel tupel_count dialog_container labelwin attrlist oneobject def_values
    #puts "fill_form OT $old_tupel $act_tupel"
    if {$old_tupel!=-1 && $act_tupel!=$old_tupel} { 
	if {![[self] update]} {
	    set act_tupel $old_tupel
	    return
	}
    }
    if {$act_tupel>=$tupel_count} {
	#Erzeuge ein leeres Prototyp Tupel
	#mit gesetzten default Werten
	foreach attr $attrlist {
	    keylset tupel $attr {}
	}
	$dialog_container set_defaults $tupel
	$dialog_container set_values $def_values
    } else {
	set tupel [[self]::myDBProxy gettupel $act_tupel]
	#puts "SET DEFAULT $tupel"
	$dialog_container set_defaults $tupel
    }
    set old_tupel $act_tupel
    if {!$oneobject} {
	$labelwin subwidget entry delete 0 end
	$labelwin subwidget entry insert 0 [expr $act_tupel +1]
    }
}
# Setzt die Verndrung von Form auf Datenbestande 
# Wenn geglckt liefert 1 zurck sonst 0
FormEngine instproc update {} {
    [self] instvar dialog_container old_tupel tupel_count def_values nestedform
    $dialog_container update_values
    set changes [$dialog_container get_changes]
    #puts "UPDATE [self] $changes"
    # Hier sollte gezeigt werden das die nderungen Ausgelesen wurden
    # wichtig beim save Funktion
    if {[lempty $changes]} { return 1 }
    if {$old_tupel>=$tupel_count} { 
	# Neues Tupel
	# berprfen der not_null Bedingung

	# Problem keine nderungen der Neu Template
	# Problem wenn alle nderungen der default entsprechen und 
	# es keine zustliche gibt dann ignorieren
	if {[keylequal def_values changes]} {
	    return 1
	}
	if {$nestedform} {
	    # Mehrfaches Vererbung oder Aggregation
	    # id wurde von dem Urvater bereit ermittelt
	    set newid [[self]::myDBProxy set refvalue]
	    [self]::myDBProxy new $changes
	} else {
	    set newid [[self]::myDBProxy new $changes]
	}
	#puts "created [self] $newid"
	[self] invokeNotify create $newid
    } else {
	# Update alter Tupel
	[self]::myDBProxy update $changes
    }
    [self] stateNeedReload
    return 1
}
# Manual Ausgefhtes Update 
FormEngine instproc updateManual {} {
    [self] instvar dialog_container
    [self] update
    $dialog_container update_defaults
}
# Diese Funktion wird nur durch den angemeldetetn CallBack von
# Vater-FormEngine aufgerufen 
FormEngine instproc createNewTupel {key} {
    #puts "create [self] $key"
    [self] instvar navwin oneobject
    [self]::myDBProxy set refvalue $key
    [self] update
}
FormEngine instproc delete {} {
    [self] instvar old_tupel tupel_count
    # Kein Tupel zum Lschen ist vorhanden
    if {$old_tupel>=$tupel_count} { 
	return
    }
    [self]::myDBProxy delete
    [self] stateNeedReload
    #puts "DELETE [self]"
    [self] invokeNotify delete
    [self] next_tupel
}
FormEngine instproc undo {} {
    [self] instvar dialog_container
    $dialog_container undo_changes
}
FormEngine instproc save_tupel {} {
    [self] update
}
FormEngine instproc filterData {} {
    [self] instvar navwin checkfilter fwin attrlist dialog_container old_filter 
    global lan
    set fwin [winfo parent $navwin].filter
    if {![winfo exist $fwin]} {
	frame $fwin  -bd 1 -relief ridge
	set checkfilter or
	radiobutton $fwin.checkor -text $lan(navigation.checkor) -value or -variable [self]::checkfilter
	radiobutton $fwin.checkand -text $lan(navigation.checkand) -value and -variable [self]::checkfilter
	tk_optionMenu $fwin.match [self]::filtermatch exact glob regexp
	button $fwin.apply -text $lan(standard.apply) -command "[self] applyFilter"
	button $fwin.drop -text $lan(standard.drop) -command "[self] dropFilter"
	pack $fwin.checkor $fwin.checkand $fwin.match $fwin.apply $fwin.drop -side left -anchor w	
    }
    if {![info exist old_filter]} {
	set old_filter {}
    }
    set fkeys [keylkeys old_filter]
    foreach attr $attrlist {
	if {![lcontain $fkeys $attr]} {
	    keylset old_filter $attr {}
	}
    }
    #  so setzten dass get_changes es erfassen kann
    $dialog_container set_values $old_filter
    pack forget $navwin 
    pack $fwin  -side bottom -anchor s -expand yes -fill x
    
}
FormEngine instproc applyFilter {} {
    [self] instvar dialog_container checkfilter filtermatch navwin fwin old_filter
    $dialog_container update_values
    set t_changes [$dialog_container get_changes]
    set changes {}
    # Leere Werte ausfiltern 
    foreach key [keylkeys t_changes] {
	if {[keylget t_changes $key]!=""} {
	    keylset changes $key [keylget t_changes $key]
	}
    }
    if {[lempty $changes]} { 
	pack forget $fwin 
	pack $navwin  -side bottom -anchor s -expand yes -fill x
	[self] fill_form
	return
    }
    set old_filter $changes
    [self]::myDBProxy setFilter $changes $checkfilter $filtermatch
    $navwin.find configure -bg blue
    pack forget $fwin 
    pack $navwin  -side bottom -anchor s -expand yes -fill x
    [self] reload
}
FormEngine instproc dropFilter {} {
    [self] instvar navwin fwin old_filter navwin
    $navwin.find configure -bg #c0c0c0
    pack forget $fwin 
    pack $navwin  -side bottom -anchor s -expand yes -fill x
    if {$old_filter!=""} {
	set old_filter {}
	[self]::myDBProxy setFilter "" 1 exact
	[self] reload
    } else {
	[self] fill_form
    }
}
FormEngine instproc check_range {} {
    [self] instvar act_tupel old_tupel tupel_count
    if {$act_tupel<0 || $act_tupel>$tupel_count} {
	set act_tupel $old_tupel
    } else {
	[self] fill_form
    }
}
FormEngine instproc next_tupel {} {
    [self] instvar act_tupel
    incr act_tupel
    [self] check_range
}
FormEngine instproc prev {} {
    [self] instvar act_tupel
    incr act_tupel -1
    [self] check_range
}
FormEngine instproc first {} {
    [self] instvar act_tupel
    set act_tupel 0
    [self] check_range
}
FormEngine instproc last {} {
    [self] instvar act_tupel tupel_count
    set act_tupel $tupel_count
    [self] check_range
}
# Forn Lschen SQL Querry beenden
FormEngine instproc destroy {} {
    [self] instvar win linkedForm oneobject sqlconnect
    if {$sqlconnect!=""} {
	# wenn es etwas nicht stimmt dann nicht aufhngen
	if {[catch {[self] update} er]!=0} {
	    global errorInfo
	    puts "update Error by [self] $er $errorInfo"
	}
    }
    #puts "destroying FormEngine"
    if {[info exist linkedForm] && [FormEngine info instances $linkedForm]!=""} {
	$linkedForm clearNotify create "[self] createNewTupel"
	# Hier kann man auch ein Kaskadierendes Delete Aufbauen"
	if {$oneobject} {
	    $linkedForm clearNotify delete "[self] delete"
	}
    }
    # zuerst die children lschen (sie brauchen die Windows noch fr update)
    foreach child [[self] info children] {
	if {[$child info class]=="::FrameContainer"} {
	    $child destroy
	}
    }
    ::destroy $win
    next
}
# Prozeduren fr die Behandlung der Formen als EditWidget
FormEngine instproc setvalue {value} {
}
FormEngine instproc updatevalue {} {
    [self] instvar linkedForm
    #puts "updatevalue [self]"
    # Ist der Formular gelinkt
    if {[info exist linkedForm] && [FormEngine info instances $linkedForm]!=""} {
	# wenn Update beim Create Tupel, dann ignoriere
	# dafr gibt es Notify createNewTupel
	if {[$linkedForm set old_tupel]>=[$linkedForm set tupel_count]} {
	    return
	}
    }
    [self] update
}
FormEngine instproc getvalue {} {
    [self] instvar value
    return $value
}
# Vorsicht!
# setdefault wird auch beim undo_changes ausgefhrt
FormEngine instproc setdefault {} {
    [self] instvar default value act_tupel tupel_count old_tupel navwin oneobject
    set value $default
    #puts "setdefault [self] VAL=$value"
    set act_tupel 0
    set old_tupel -1
    # value=="" soll als neues Tupel implementiert werden
    #if {$value==""} return
    if {$default==""} {
	set tupel_count 0
	[self] fill_form
	if {$oneobject} {
	    $navwin.delete configure -state disabled
	    $navwin.reload configure -state disabled
	    $navwin.save configure -state disabled
	} else {
	    $navwin.first configure -state disabled
	    $navwin.prev configure -state disabled
	    $navwin.next configure -state disabled
	    $navwin.last configure -state disabled
	    $navwin.delete configure -state disabled
	    $navwin.reload configure -state disabled
	    $navwin.find configure -state disabled
	    $navwin.save configure -state disabled
	}
    } else {
	[self]::myDBProxy set refvalue $default
	[self] reload
    }
}

# DBProxy ist ein Vermitler zwischen SQL Verbindung und Engine
# Er ist spezielisiert Tabelen abzufragen und Ergebnisse nach Index zu Liefern
# auch Tupel Anzahl mu er speichern
# Er wird bentigt um vielleicht Ergebnisse zu puffern
# Die unterschiedliche Arten von Abfragen zu kapseln
# oder andere als SQl Interfaces zu realisieren
#
# Hier kann auch die Zentrale Stelle sein um neue oder modefizierte 
# Tupel zu merken
Class DBProxy -parameter {sqlconnect table attrlist primarykey filter refvalue {refattribute {}}}
DBProxy instproc init args {
    [self] instvar attrlist primarykey  qdata refattribute qfilter
    if {![lcontain $attrlist $refattribute] && $refattribute!=""} {
	lappend attrlist $refattribute
    }
    set notdata [minusMenge $primarykey $attrlist]
    if {[lempty $notdata]} {
	set qdata $attrlist
    } else {
	set qdata [concat $attrlist $notdata]
    }
    set qfilter {}
}
DBProxy instproc reload {} {
    [self] instvar sqlconnect table attrlist qdata primarykey qfilter query lastfetch refattribute refvalue
    if {[info exist [self]::query]} {
	$query destroy
    }
    set sqlstatment "SELECT [join $qdata ,] FROM $table"
    set where ""
    if {$refattribute!=""} {
	append where " WHERE $refattribute=$refvalue"
    }
    if {$qfilter!=""} {
	if {$where==""} {
	    append where " WHERE $qfilter"
	} else {
	    append where " AND ( $qfilter )"
	}
    }
    append sqlstatment $where
    #puts "QUERY $sqlstatment"
    set query [$sqlconnect query $sqlstatment]
    set lastfetch -1
    return [$query rows]
}
# Nimmt ein Tupel als tclx keyedlist
# lastfetch ist ein ID des letzten genommenen Tuppels
# es wird immer die Identifikation des letzten Tuppel gespeichert
# als keyedlist tupelid
# um in nacher modifizieren oder lschen zu knnen
DBProxy instproc gettupel {id} {
    [self] instvar query lastfetch attrlist qdata tupelid primarykey
    if {$id==[expr $lastfetch+1]} {
	# Einfach der nschtr
	set row [$query fetch]
	incr lastfetch
    } else {
	#Pointer setzten
	$query seek $id
	set lastfetch $id
	set row [$query fetch]
    }
    #puts "GET $row"
    set tupel {}
    set index 0
    foreach column $attrlist {
	keylset tupel $column [lindex $row $index]
	incr index
    }
    foreach keyattr $primarykey {
	keylset tupelid $keyattr [lindex $row [lsearch $qdata $keyattr]]
    }
    return $tupel
}
# Es werden Zeichenketten als SQL Strings maskiert
# Ersetzung von " und \
# und Umschlieen mit "............"
# NULL wird nicht maskiert
DBProxy instproc maskstring {string} {
    if {$string=="NULL"} {return NULL}
    regsub -all \" $string \\\" string
    regsub -all {\\} $string {\\\\} string
    return "'$string'"  
}
# Baut die richtige WHERE Klausel aus tupelid (interne Benutzung)
DBProxy instproc buildwhere {} {
    [self] instvar tupelid refattribute refvalue
    set where {}
    foreach attr [keylkeys tupelid] {
	if {$where!=""} { append where " AND " }
	if {[set val  [keylget tupelid $attr]]!="NULL"} {
	    append where "$attr=[[self] maskstring $val]"
	} else {
	    append where "$attr IS NULL"
	}
    }
    # Bei Vererbung kann ein refattribute zum prim gehren
    # also keine verdopplung
    if {$refattribute!="" && ![lcontain [keylkeys tupelid] $refattribute]} {
	append where " AND $refattribute=$refvalue"
    }
    return $where
}
# Baut einen Filter (oder Suchfunktion) ber die Tupel
# filterattr - ist eine Keydlist (TclX) ber die Attributen
# boolmode - kann "and" oder "or" sein (komplexe Abfragen so nicht mglich)
# filtermatch - kann "exact" "glob" oder "regexp" sein die Abfrage wird 
#               gebildet mit = Like und =~
# Die einzelnen Attributfilter knnen modifikatoren haben
# = > < >= <= ! (not)  ~ (regexp) * (glob)
# \ maskiert den Modifiekator
DBProxy instproc setFilter {filterattr boolmode filtermatch} {
    [self] instvar qfilter
    set qfilter {}
    foreach attr [keylkeys filterattr] {
	if {$qfilter!=""} {
	    if {$boolmode=="and"} {
		append qfilter " AND "
	    } else {
		append qfilter " OR "
	    }
	}
	set val [keylget filterattr $attr]
	set mod [string range $val 0 0]
	set op {}
	set noval 0
	switch -- $mod {
	    = -
	    > -
	    < {
		set val [string range $val 1 end]
		set op $mod
	    }
	    * {
		set val [string range $val 1 end]
		set op LIKE
	    }
	    ~ {
		set val [string range $val 1 end]
		set op REGEXP
	    }
	    ! {
		switch [string range $val 1 1] {
		    = {
			set val [string range $val 2 end]
			set op !=
		    }
		    * {
			set val [string range $val 2 end]
			set op "NOT LIKE"
		    }
		    ~ {
			set val [string range $val 2 end]
			set op "NOT REGEXP"
		    }
		    default {
			set val [string range $val 1 end]
			set op !=
		    }
		}
	    }
	    \\ {
		set val [string range $val 1 end]
	    }
	    default {
		switch $val {
		    "IS NULL" {
			set op "IS NULL"
			set noval 1
		    }
		    "IS NOT NULL" {
			set op "IS NOT NULL"
			set noval 1
		    }
		}
	    }
	}
	#puts "OP $op mod=$mod"
	if {$op==""} {
	    switch $filtermatch {
		glob {
		    set op LIKE
		}
		regexp {
		    set op REGEXP
		}
		exact -
		default {
		    set op =
		}
	    }
	}
	if {$noval} {
	    append qfilter "$attr $op"
	} else {
	    append qfilter "$attr $op [[self] maskstring $val]"
	}
    }
    #puts "setFilter $filterattr $qfilter"
}
# Fht die nderungen an dem letzten gelesenen Tupel
DBProxy instproc update {changes} {
    [self] instvar sqlconnect table
    set updateset {}
    foreach attr [keylkeys changes] {
	if {$updateset!=""} { append updateset "," }
	append updateset "$attr=[[self] maskstring [keylget changes $attr]]"
    }
    $sqlconnect execute "UPDATE $table SET $updateset WHERE [[self] buildwhere]"
}
DBProxy instproc delete {} {
    [self] instvar sqlconnect table
    $sqlconnect execute "DELETE FROM $table WHERE [[self] buildwhere]"
}
# Gibt zurck den primrschlssel des neuen Tupel fall aus einen Attribut
# sonst NOPRIM
DBProxy instproc new {values} {
    [self] instvar sqlconnect table refattribute refvalue primarykey
    set newvalues {}
    foreach attr [keylkeys values] {
	if {$newvalues!=""} { append newvalues ","; append collist "," }
	append newvalues [maskstring [keylget values $attr]]
	append collist $attr
    }
    if {$refattribute!=""} {
	append newvalues ",$refvalue"
	append collist ",$refattribute"
    }
    $sqlconnect execute "INSERT INTO $table ($collist) VALUES ($newvalues)"
    if {[llength $primarykey]!=1} {
	return NOPRIM
    } else {
	if {[lcontain [keylkeys values] $primarykey]} {
	    return [keylget values $primarykey]
	} else {
	    return [$sqlconnect getinsertid]
	}
    }
}
DBProxy instproc destroy {} {
    [self] instvar  query 
    $query destroy
    #puts "destroy DBProxy [self]"
    next
}
Class FormLink  -superclass InputWidget -parameter {
    form_node
    sqlconnect 
    data
    label
    {refattribute {}}
}
FormLink metadata add attributes
FormLink metadata attributes {label form refattribute data}
#
# Realiseirung von Form Navigation (FormLink)
# Das Eigentliche Form wird in das Objekt Aggregiert  
# die nderungen werden durch 
FormLink instproc map {window} {
    [self] instvar win label
    set win $window
    button $win -text $label -command "[self] executeForm"
    return $win
}
FormLink instproc executeForm {} {
    [self] instvar form_node sqlconnect data refattribute formengine
    if {![info exist formengine] || \
	     $formengine=="" || \
	     [FormEngine info instances ::$formengine]==""} {
	# suche den Vatet FormEngine
	# FormLink soll der Kind oder Enkelkind .... davon sein
	for {set pf_engine [[self] info parent]} \
	    {$pf_engine!="" && $pf_engine!="::" && [$pf_engine info class]!="::FormEngine"} \
	    {set pf_engine [$pf_engine info parent]} {#puts "$pf_engine [$pf_engine info class]"}
	set formengine [FormEngine [Object autoname formengine] \
			    -form_node $form_node \
			    -sqlconnect $sqlconnect -data $data \
			    -refattribute $refattribute -linkedForm $pf_engine
		       ]
	$formengine map
	[self] setdefault
    }
}
# Prozeduren fr die Behandlung der FormLink als EditWidget
FormLink instproc setvalue {value} {
}
FormLink instproc updatevalue {} {
    [self] instvar formengine
    if {[info exist formengine] &&  $formengine!="" &&  [FormEngine info instances ::$formengine]!=""} {
	$formengine updatevalue
    }
}
FormLink instproc getvalue {} {
    [self] instvar value
    return $value
}
FormLink instproc setdefault {} {
    [self] instvar default value formengine sqlconnect
    if {$sqlconnect==""} return
    set value $default
    if {[info exist formengine] &&  $formengine!="" &&  [FormEngine info instances ::$formengine]!=""} {
	$formengine set default $default
	$formengine setdefault
    }
}
# Die Vererbung wird durch ein NoteBook Widget wiedergeben.
# Dabei werden die unterschiedliche Argumente des 
# Kindobjekt als Unterformulare (Eigebetette Formulare) gezeigt
#
# Das Objekt besitzt entweder aggregiertes ONoteBookConteiner Objekt mit
# einzelnen Unterformularen(FormEngine)  (notebook) oder eine aggregierte 
# FormEngine als (form)
#
# ! Die Objekte der Klasse knnen sich selber nicht kontruieren
Class InheritForm  -superclass InputWidget 
InheritForm metadata add attributes
InheritForm metadata attributes {data}
InheritForm instproc map {window} {
    [self] instvar win
    set win $window
    if {[[self] info children notebook]!=""} {
	[self]::notebook map $window
    } else {
	[self]::form map $window
    }
    return $win
}
InheritForm instproc setvalue {value} {
}
InheritForm instproc getAllChildren {} {
    if {[[self] info children notebook]!=""} {
	return  [[self]::notebook info children]
    } else {
	return [self]::form
    }
}
InheritForm instproc updatevalue {} {
    [self] instvar data
    foreach child [[self] getAllChildren] {
	    $child updatevalue
    }
}
InheritForm instproc getvalue {} {
    [self] instvar value
    return $value
}
InheritForm instproc setdefault {} {
    [self] instvar default value data 
    set value $default
    # Eigentlich ein Neues Template
    set foundchild 0
    foreach child [[self] getAllChildren] {
	#puts "seting $child"
	$child set default $default
	$child setdefault
	#[self]::notebook::${echild}::myDBProxy set refvalue $default
	#[self]::notebook::${echild} reload
	# Schalte auf das NoteBook wenn Kind gefunden
	if {!$foundchild && [[self] info children notebook]!=""} {
	    if {[$child set tupel_count]>0} {
		[self]::notebook raise [string tolower [namespace tail $child]]
		set foundchild 1
	    }
	}
	
    }
}
